@@ -1,24 +1,49 @@
-import os
-import tempfile
+import copy
 import json
-import setpath  # noqa:F401, must come before 'import mechanicalsoup'
-import mechanicalsoup
-import sys
+import os
 import re
-from bs4 import BeautifulSoup
-from utils import setup_mock_browser, prepare_mock_browser, mock_get
-import pytest
+import sys
+import tempfile
 import webbrowser
 
+import pytest
+import setpath  # noqa:F401, must come before 'import mechanicalsoup'
+from bs4 import BeautifulSoup
+from utils import (mock_get, open_legacy_httpbin, prepare_mock_browser,
+                   setup_mock_browser)
+
+import mechanicalsoup
+import requests
+
 
 def test_request_forward():
-    browser, url = setup_mock_browser(expected_post=[('var1', 'val1'),
-                                                     ('var2', 'val2')])
-    r = browser.request('POST', url + '/post', data={'var1': 'val1',
-                                                     'var2': 'val2'})
+    data = [('var1', 'val1'), ('var2', 'val2')]
+    browser, url = setup_mock_browser(expected_post=data)
+    r = browser.request('POST', url + '/post', data=data)
     assert r.text == 'Success!'
 
 
+def test_properties():
+    """Check that properties return the same value as the getter."""
+    browser = mechanicalsoup.StatefulBrowser()
+    browser.open_fake_page('<form></form>', url="http://example.com")
+    assert browser.page == browser.get_current_page()
+    assert browser.page is not None
+    assert browser.url == browser.get_url()
+    assert browser.url is not None
+    browser.select_form()
+    assert browser.form == browser.get_current_form()
+    assert browser.form is not None
+
+
+def test_get_selected_form_unselected():
+    browser = mechanicalsoup.StatefulBrowser()
+    browser.open_fake_page('<form></form>')
+    with pytest.raises(AttributeError, match="No form has been selected yet."):
+        browser.form
+    assert browser.get_current_form() is None
+
+
 def test_submit_online(httpbin):
     """Complete and submit the pizza form at http://httpbin.org/forms/post """
     browser = mechanicalsoup.StatefulBrowser()
@@ -29,7 +54,7 @@ def test_submit_online(httpbin):
             browser.follow_link(link)
             break
     browser.follow_link("forms/post")
-    assert browser.get_url() == httpbin + "/forms/post"
+    assert browser.url == httpbin + "/forms/post"
     browser.select_form("form")
     browser["custname"] = "Customer Name Here"
     browser["size"] = "medium"
@@ -37,14 +62,14 @@ def test_submit_online(httpbin):
     # Change our mind to make sure old boxes are unticked
     browser["topping"] = ("cheese", "onion")
     browser["comments"] = "Some comment here"
-    browser.get_current_form().set("nosuchfield", "new value", True)
+    browser.form.set("nosuchfield", "new value", True)
     response = browser.submit_selected()
     json = response.json()
     data = json["form"]
     assert data["custname"] == "Customer Name Here"
     assert data["custtel"] == ""  # web browser submits "" for input left blank
     assert data["size"] == "medium"
-    assert set(data["topping"]) == set(("cheese", "onion"))
+    assert set(data["topping"]) == {"cheese", "onion"}
     assert data["comments"] == "Some comment here"
     assert data["nosuchfield"] == "new value"
 
@@ -64,7 +89,7 @@ def test_no_404(httpbin):
 def test_404(httpbin):
     browser = mechanicalsoup.StatefulBrowser(raise_on_404=True)
     with pytest.raises(mechanicalsoup.LinkNotFoundError):
-        resp = browser.open(httpbin + "/nosuchpage")
+        browser.open(httpbin + "/nosuchpage")
     resp = browser.open(httpbin.url)
     assert resp.status_code == 200
 
@@ -83,11 +108,11 @@ def test_open_relative(httpbin):
     # Open a relative page and make sure remote host and browser agree on URL
     resp = browser.open_relative("/get")
     assert resp.json()['url'] == httpbin + "/get"
-    assert browser.get_url() == httpbin + "/get"
+    assert browser.url == httpbin + "/get"
 
     # Test passing additional kwargs to the session
     resp = browser.open_relative("/basic-auth/me/123", auth=('me', '123'))
-    assert browser.get_url() == httpbin + "/basic-auth/me/123"
+    assert browser.url == httpbin + "/basic-auth/me/123"
     assert resp.json() == {"authenticated": True, "user": "me"}
 
 
@@ -121,15 +146,15 @@ def test_links():
 @pytest.mark.parametrize("expected_post", [
     pytest.param(
         [
+            ('text', 'Setting some text!'),
             ('comment', 'Selecting an input submit'),
             ('diff', 'Review Changes'),
-            ('text', 'Setting some text!')
         ], id='input'),
     pytest.param(
         [
+            ('text', '= Heading =\n\nNew page here!\n'),
             ('comment', 'Selecting a button submit'),
             ('cancel', 'Cancel'),
-            ('text', '= Heading =\n\nNew page here!\n')
         ], id='button'),
 ])
 def test_submit_btnName(expected_post):
@@ -137,18 +162,74 @@ def test_submit_btnName(expected_post):
     browser, url = setup_mock_browser(expected_post=expected_post)
     browser.open(url)
     browser.select_form('#choose-submit-form')
-    browser['text'] = expected_post[2][1]
-    browser['comment'] = expected_post[0][1]
-    res = browser.submit_selected(btnName=expected_post[1][0])
-    assert(res.status_code == 200 and res.text == 'Success!')
+    browser['text'] = dict(expected_post)['text']
+    browser['comment'] = dict(expected_post)['comment']
+    initial_state = browser._StatefulBrowser__state
+    res = browser.submit_selected(btnName=expected_post[2][0])
+    assert res.status_code == 200 and res.text == 'Success!'
+    assert initial_state != browser._StatefulBrowser__state
+
+
+@pytest.mark.parametrize("expected_post", [
+    pytest.param(
+        [
+            ('text', 'Setting some text!'),
+            ('comment', 'Selecting an input submit'),
+        ], id='input'),
+    pytest.param(
+        [
+            ('text', '= Heading =\n\nNew page here!\n'),
+            ('comment', 'Selecting a button submit'),
+        ], id='button'),
+])
+def test_submit_no_btn(expected_post):
+    '''Tests that no submit inputs are posted when btnName=False.'''
+    browser, url = setup_mock_browser(expected_post=expected_post)
+    browser.open(url)
+    browser.select_form('#choose-submit-form')
+    browser['text'] = dict(expected_post)['text']
+    browser['comment'] = dict(expected_post)['comment']
+    initial_state = browser._StatefulBrowser__state
+    res = browser.submit_selected(btnName=False)
+    assert res.status_code == 200 and res.text == 'Success!'
+    assert initial_state != browser._StatefulBrowser__state
+
+
+def test_submit_dont_modify_kwargs():
+    """Test that submit_selected() doesn't modify the caller's passed-in
+    kwargs, for example when adding a Referer header.
+    """
+    kwargs = {'headers': {'Content-Type': 'text/html'}}
+    saved_kwargs = copy.deepcopy(kwargs)
+
+    browser, url = setup_mock_browser(expected_post=[], text='<form></form>')
+    browser.open(url)
+    browser.select_form()
+    browser.submit_selected(**kwargs)
+
+    assert kwargs == saved_kwargs
+
+
+def test_submit_dont_update_state():
+    expected_post = [
+            ('text', 'Bananas are good.'),
+            ('preview', 'Preview Page')]
+    browser, url = setup_mock_browser(expected_post=expected_post)
+    browser.open(url)
+    browser.select_form('#choose-submit-form')
+    browser['text'] = dict(expected_post)['text']
+    initial_state = browser._StatefulBrowser__state
+    res = browser.submit_selected(update_state=False)
+    assert res.status_code == 200 and res.text == 'Success!'
+    assert initial_state == browser._StatefulBrowser__state
 
 
 def test_get_set_debug():
     browser = mechanicalsoup.StatefulBrowser()
     # Debug mode is off by default
-    assert(not browser.get_debug())
+    assert not browser.get_debug()
     browser.set_debug(True)
-    assert(browser.get_debug())
+    assert browser.get_debug()
 
 
 def test_list_links(capsys):
@@ -158,10 +239,10 @@ def test_list_links(capsys):
      <a href="/link1">Link #1</a>
      <a href="/link2" id="link2"> Link #2</a>
 '''
-    browser.open_fake_page('<html>{0}</html>'.format(links))
+    browser.open_fake_page(f'<html>{links}</html>')
     browser.list_links()
     out, err = capsys.readouterr()
-    expected = 'Links in the current page:{0}'.format(links)
+    expected = f'Links in the current page:{links}'
     assert out == expected
 
 
@@ -210,9 +291,9 @@ def test_verbose(capsys):
     assert browser.get_verbose() == 2
 
 
-def test_new_control():
+def test_new_control(httpbin):
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open("http://httpbin.org/forms/post")
+    browser.open(httpbin + "/forms/post")
     browser.select_form("form")
     with pytest.raises(mechanicalsoup.LinkNotFoundError):
         # The control doesn't exist, yet.
@@ -222,8 +303,11 @@ def test_new_control():
     browser.new_control("text", "temperature", "warm")
     browser.new_control("textarea", "size", "Sooo big !")
     browser.new_control("text", "comments", "This is an override comment")
+    fake_select = BeautifulSoup("", "html.parser").new_tag('select')
+    fake_select["name"] = "foo"
+    browser.form.form.append(fake_select)
     browser.new_control("checkbox", "foo", "valval", checked="checked")
-    tag = browser.get_current_form().form.find("input", {"name": "foo"})
+    tag = browser.form.form.find("input", {"name": "foo"})
     assert tag.attrs["checked"] == "checked"
     browser["temperature"] = "hot"
     response = browser.submit_selected()
@@ -253,7 +337,7 @@ def test_form_noaction():
     browser, url = setup_mock_browser()
     browser.open_fake_page(submit_form_noaction)
     browser.select_form('#choose-submit-form')
-    with pytest.raises(ValueError, message="no URL to submit to"):
+    with pytest.raises(ValueError, match="no URL to submit to"):
         browser.submit_selected()
 
 
@@ -278,7 +362,7 @@ def test_form_noname():
     browser.open_fake_page(submit_form_noname, url=url)
     browser.select_form('#choose-submit-form')
     response = browser.submit_selected()
-    assert(response.status_code == 200 and response.text == 'Success!')
+    assert response.status_code == 200 and response.text == 'Success!'
 
 
 submit_form_multiple = '''
@@ -297,40 +381,87 @@ submit_form_multiple = '''
 
 
 def test_form_multiple():
-    browser, url = setup_mock_browser(expected_post=[('foo', 'tempeh'),
-                                                     ('foo', 'tofu')])
+    browser, url = setup_mock_browser(expected_post=[('foo', 'tofu'),
+                                                     ('foo', 'tempeh')])
     browser.open_fake_page(submit_form_multiple, url=url)
     browser.select_form('#choose-submit-form')
     response = browser.submit_selected()
-    assert(response.status_code == 200 and response.text == 'Success!')
+    assert response.status_code == 200 and response.text == 'Success!'
 
 
 def test_upload_file(httpbin):
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open(httpbin + "/forms/post")
+    url = httpbin + "/post"
+    file_input_form = f"""
+    <form method="post" action="{url}" enctype="multipart/form-data">
+        <input type="file" name="first" />
+    </form>
+    """
 
     # Create two temporary files to upload
     def make_file(content):
         path = tempfile.mkstemp()[1]
-        with open(path, "w") as f:
-            f.write(content)
+        with open(path, "w") as fd:
+            fd.write(content)
         return path
-    path1, path2 = (make_file(content) for content in
-                    ("first file content", "second file content"))
+    path1 = make_file("first file content")
+    path2 = make_file("second file content")
 
-    # The form doesn't have a type=file field, but the target action
-    # does show it => add the fields ourselves.
+    value1 = open(path1, "rb")
+    value2 = open(path2, "rb")
+
+    browser.open_fake_page(file_input_form)
     browser.select_form()
-    browser.new_control("file", "first", path1)
-    browser.new_control("file", "second", "")
-    browser["second"] = path2
-    browser.get_current_form().print_summary()
+
+    # Test filling an existing input and creating a new input
+    browser["first"] = value1
+    browser.new_control("file", "second", value2)
+
     response = browser.submit_selected()
     files = response.json()["files"]
     assert files["first"] == "first file content"
     assert files["second"] == "second file content"
 
 
+def test_upload_file_with_malicious_default(httpbin):
+    """Check for CVE-2023-34457 by setting the form input value directly to a
+    file that the user does not explicitly consent to upload, as a malicious
+    server might do.
+    """
+    browser = mechanicalsoup.StatefulBrowser()
+    sensitive_path = tempfile.mkstemp()[1]
+    with open(sensitive_path, "w") as fd:
+        fd.write("Some sensitive information")
+    url = httpbin + "/post"
+    malicious_html = f"""
+    <form method="post" action="{url}" enctype="multipart/form-data">
+        <input type="file" name="malicious" value="{sensitive_path}" />
+    </form>
+    """
+    browser.open_fake_page(malicious_html)
+    browser.select_form()
+    response = browser.submit_selected()
+    assert response.json()["files"] == {"malicious": ""}
+
+
+def test_upload_file_raise_on_string_input():
+    """Check for use of the file upload API that was modified to remediate
+    CVE-2023-34457. Users must now open files manually to upload them.
+    """
+    browser = mechanicalsoup.StatefulBrowser()
+    file_input_form = """
+    <form enctype="multipart/form-data">
+        <input type="file" name="upload" />
+    </form>
+    """
+    browser.open_fake_page(file_input_form)
+    browser.select_form()
+    with pytest.raises(ValueError, match="CVE-2023-34457"):
+        browser["upload"] = "/path/to/file"
+    with pytest.raises(ValueError, match="CVE-2023-34457"):
+        browser.new_control("file", "upload2", "/path/to/file")
+
+
 def test_with():
     """Test that __enter__/__exit__ properly create/close the browser."""
     with mechanicalsoup.StatefulBrowser() as browser:
@@ -365,13 +496,38 @@ def test_select_form_tag_object():
             browser.select_form(soup.find("p"))
 
 
+def test_select_form_associated_elements():
+    """Test associated elements outside the form tag"""
+    forms = """<form id="a"><input><textarea></form><input form="a">
+               <textarea form="a"/><input form="b">
+               <form id="ab" action="/test.php"><input></form>
+               <textarea form="ab"></textarea>
+            """
+    with mechanicalsoup.StatefulBrowser() as browser:
+        browser.open_fake_page(forms)
+        elements_form_a = set([
+            "<input/>", "<textarea></textarea>",
+            '<input form="a"/>', '<textarea form="a"></textarea>'])
+        elements_form_ab = set(["<input/>", '<textarea form="ab"></textarea>'])
+        form_by_str = browser.select_form("#a")
+        form_by_tag = browser.select_form(browser.page.find("form", id='a'))
+        form_by_css = browser.select_form("form[action$='.php']")
+        assert set([str(element) for element in form_by_str.form.find_all((
+            "input", "textarea"))]) == elements_form_a
+        assert set([str(element) for element in form_by_tag.form.find_all((
+            "input", "textarea"))]) == elements_form_a
+        assert set([str(element) for element in form_by_css.form.find_all((
+            "input", "textarea"))]) == elements_form_ab
+
+
 def test_referer_follow_link(httpbin):
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open(httpbin.url)
+    open_legacy_httpbin(browser, httpbin)
+    start_url = browser.url
     response = browser.follow_link("/headers")
     referer = response.json()["headers"]["Referer"]
     actual_ref = re.sub('/*$', '', referer)
-    expected_ref = re.sub('/*$', '', httpbin.url)
+    expected_ref = re.sub('/*$', '', start_url)
     assert actual_ref == expected_ref
 
 
@@ -401,6 +557,26 @@ def test_referer_submit(httpbin):
     assert actual_ref == ref
 
 
+@pytest.mark.parametrize("referer_header", ["Referer", "referer"])
+def test_referer_submit_override(httpbin, referer_header):
+    """Ensure the caller can override the Referer header that
+    mechanicalsoup would normally add. Because headers are case insensitive,
+    test with both 'Referer' and 'referer'.
+    """
+
+    browser = mechanicalsoup.StatefulBrowser()
+    ref = "https://example.com/my-referer"
+    ref_override = "https://example.com/override"
+    page = submit_form_headers.format(httpbin.url + "/headers")
+    browser.open_fake_page(page, url=ref)
+    browser.select_form()
+    response = browser.submit_selected(headers={referer_header: ref_override})
+    headers = response.json()["headers"]
+    referer = headers["Referer"]
+    actual_ref = re.sub('/*$', '', referer)
+    assert actual_ref == ref_override
+
+
 def test_referer_submit_headers(httpbin):
     browser = mechanicalsoup.StatefulBrowser()
     ref = "https://example.com/my-referer"
@@ -416,45 +592,74 @@ def test_referer_submit_headers(httpbin):
     assert headers['X-Test-Header'] == 'x-test-value'
 
 
-def test_link_arg_text(httpbin):
+@pytest.mark.parametrize('expected, kwargs', [
+    pytest.param('/foo', {}, id='none'),
+    pytest.param('/get', {'string': 'Link'}, id='string'),
+    pytest.param('/get', {'url_regex': 'get'}, id='regex'),
+])
+def test_follow_link_arg(httpbin, expected, kwargs):
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open_fake_page('<a href="/get">Link</a>', httpbin.url)
-    browser.follow_link(link_text='Link')
-    assert browser.get_url() == httpbin + '/get'
+    html = '<a href="/foo">Bar</a><a href="/get">Link</a>'
+    browser.open_fake_page(html, httpbin.url)
+    browser.follow_link(bs4_kwargs=kwargs)
+    assert browser.url == httpbin + expected
 
 
-def test_link_arg_regex(httpbin):
+def test_follow_link_excess(httpbin):
+    """Ensure that excess args are passed to BeautifulSoup"""
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open_fake_page('<a href="/get">Link</a>', httpbin.url)
-    browser.follow_link(url_regex='.*')
-    assert browser.get_url() == httpbin + '/get'
+    html = '<a href="/foo">Bar</a><a href="/get">Link</a>'
+    browser.open_fake_page(html, httpbin.url)
+    browser.follow_link(url_regex='get')
+    assert browser.url == httpbin + '/get'
 
-
-def test_link_arg_multiregex(httpbin):
     browser = mechanicalsoup.StatefulBrowser()
     browser.open_fake_page('<a href="/get">Link</a>', httpbin.url)
     with pytest.raises(ValueError, match="link parameter cannot be .*"):
         browser.follow_link('foo', url_regex='bar')
 
 
+def test_follow_link_ua(httpbin):
+    """Tests passing requests parameters to follow_link() by
+    setting the User-Agent field."""
+    browser = mechanicalsoup.StatefulBrowser()
+    # html = '<a href="/foo">Bar</a><a href="/get">Link</a>'
+    # browser.open_fake_page(html, httpbin.url)
+    open_legacy_httpbin(browser, httpbin)
+    bs4_kwargs = {'url_regex': 'user-agent'}
+    requests_kwargs = {'headers': {"User-Agent": '007'}}
+    resp = browser.follow_link(bs4_kwargs=bs4_kwargs,
+                               requests_kwargs=requests_kwargs)
+    assert browser.url == httpbin + '/user-agent'
+    assert resp.json() == {'user-agent': '007'}
+    assert resp.request.headers['user-agent'] == '007'
+
+
+def test_link_arg_multiregex(httpbin):
+    browser = mechanicalsoup.StatefulBrowser()
+    browser.open_fake_page('<a href="/get">Link</a>', httpbin.url)
+    with pytest.raises(ValueError, match="link parameter cannot be .*"):
+        browser.follow_link('foo', bs4_kwargs={'url_regex': 'bar'})
+
+
 def file_get_contents(filename):
-    with open(filename, "rb") as f:
-        return f.read()
+    with open(filename, "rb") as fd:
+        return fd.read()
 
 
 def test_download_link(httpbin):
     """Test downloading the contents of a link to file."""
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open(httpbin.url)
+    open_legacy_httpbin(browser, httpbin)
     tmpdir = tempfile.mkdtemp()
     tmpfile = tmpdir + '/nosuchfile.png'
-    current_url = browser.get_url()
-    current_page = browser.get_current_page()
+    current_url = browser.url
+    current_page = browser.page
     response = browser.download_link(file=tmpfile, link='image/png')
 
     # Check that the browser state has not changed
-    assert browser.get_url() == current_url
-    assert browser.get_current_page() == current_page
+    assert browser.url == current_url
+    assert browser.page == current_page
 
     # Check that the file was downloaded
     assert os.path.isfile(tmpfile)
@@ -466,34 +671,86 @@ def test_download_link(httpbin):
 def test_download_link_nofile(httpbin):
     """Test downloading the contents of a link without saving it."""
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open(httpbin.url)
-    current_url = browser.get_url()
-    current_page = browser.get_current_page()
+    open_legacy_httpbin(browser, httpbin)
+    current_url = browser.url
+    current_page = browser.page
     response = browser.download_link(link='image/png')
 
     # Check that the browser state has not changed
-    assert browser.get_url() == current_url
-    assert browser.get_current_page() == current_page
+    assert browser.url == current_url
+    assert browser.page == current_page
 
     # Check that we actually downloaded a PNG file
     assert response.content[:4] == b'\x89PNG'
 
 
+def test_download_link_nofile_bs4(httpbin):
+    """Test downloading the contents of a link without saving it."""
+    browser = mechanicalsoup.StatefulBrowser()
+    open_legacy_httpbin(browser, httpbin)
+    current_url = browser.url
+    current_page = browser.page
+    response = browser.download_link(bs4_kwargs={'url_regex': 'image.png'})
+
+    # Check that the browser state has not changed
+    assert browser.url == current_url
+    assert browser.page == current_page
+
+    # Check that we actually downloaded a PNG file
+    assert response.content[:4] == b'\x89PNG'
+
+
+def test_download_link_nofile_excess(httpbin):
+    """Test downloading the contents of a link without saving it."""
+    browser = mechanicalsoup.StatefulBrowser()
+    open_legacy_httpbin(browser, httpbin)
+    current_url = browser.url
+    current_page = browser.page
+    response = browser.download_link(url_regex='image.png')
+
+    # Check that the browser state has not changed
+    assert browser.url == current_url
+    assert browser.page == current_page
+
+    # Check that we actually downloaded a PNG file
+    assert response.content[:4] == b'\x89PNG'
+
+
+def test_download_link_nofile_ua(httpbin):
+    """Test downloading the contents of a link without saving it."""
+    browser = mechanicalsoup.StatefulBrowser()
+    open_legacy_httpbin(browser, httpbin)
+    current_url = browser.url
+    current_page = browser.page
+    requests_kwargs = {'headers': {"User-Agent": '007'}}
+    response = browser.download_link(link='image/png',
+                                     requests_kwargs=requests_kwargs)
+    # Check that the browser state has not changed
+    assert browser.url == current_url
+    assert browser.page == current_page
+
+    # Check that we actually downloaded a PNG file
+    assert response.content[:4] == b'\x89PNG'
+
+    # Check that we actually set the User-agent outbound
+    assert response.request.headers['user-agent'] == '007'
+
+
 def test_download_link_to_existing_file(httpbin):
     """Test downloading the contents of a link to an existing file."""
     browser = mechanicalsoup.StatefulBrowser()
-    browser.open(httpbin.url)
+    open_legacy_httpbin(browser, httpbin)
     tmpdir = tempfile.mkdtemp()
     tmpfile = tmpdir + '/existing.png'
-    with open(tmpfile, "w") as f:
-        f.write("initial content")
-    current_url = browser.get_url()
-    current_page = browser.get_current_page()
+    with open(tmpfile, "w") as fd:
+        fd.write("initial content")
+    current_url = browser.url
+    current_page = browser.page
     response = browser.download_link('image/png', tmpfile)
 
     # Check that the browser state has not changed
-    assert browser.get_url() == current_url
-    assert browser.get_current_page() == current_page
+    assert browser.url == current_url
+    assert browser.page == current_page
 
     # Check that the file was downloaded
     assert os.path.isfile(tmpfile)
@@ -509,14 +766,14 @@ def test_download_link_404(httpbin):
                            url=httpbin.url)
     tmpdir = tempfile.mkdtemp()
     tmpfile = tmpdir + '/nosuchfile.txt'
-    current_url = browser.get_url()
-    current_page = browser.get_current_page()
+    current_url = browser.url
+    current_page = browser.page
     with pytest.raises(mechanicalsoup.LinkNotFoundError):
         browser.download_link(file=tmpfile, link_text='Link')
 
     # Check that the browser state has not changed
-    assert browser.get_url() == current_url
-    assert browser.get_current_page() == current_page
+    assert browser.url == current_url
+    assert browser.page == current_page
 
     # Check that the file was not downloaded
     assert not os.path.exists(tmpfile)
@@ -529,17 +786,17 @@ def test_download_link_referer(httpbin):
     browser.open_fake_page('<a href="/headers">Link</a>',
                            url=ref)
     tmpfile = tempfile.NamedTemporaryFile()
-    current_url = browser.get_url()
-    current_page = browser.get_current_page()
+    current_url = browser.url
+    current_page = browser.page
     browser.download_link(file=tmpfile.name, link_text='Link')
 
     # Check that the browser state has not changed
-    assert browser.get_url() == current_url
-    assert browser.get_current_page() == current_page
+    assert browser.url == current_url
+    assert browser.page == current_page
 
     # Check that the file was downloaded
-    with open(tmpfile.name) as f:
-        json_data = json.load(f)
+    with open(tmpfile.name) as fd:
+        json_data = json.load(fd)
     headers = json_data["headers"]
     assert headers["Referer"] == ref
 
@@ -557,14 +814,14 @@ def test_refresh_open():
 
     browser.refresh()
 
-    assert browser.get_url() == url
-    assert browser.get_current_page() == reload_page
+    assert browser.url == url
+    assert browser.page == reload_page
 
 
 def test_refresh_follow_link():
     url = 'mock://example.com'
     follow_url = 'mock://example.com/followed'
-    initial_content = '<a href="{url}">Link</a>'.format(url=follow_url)
+    initial_content = f'<a href="{follow_url}">Link</a>'
     initial_page = BeautifulSoup(initial_content, 'lxml')
     reload_page = BeautifulSoup('<p>Fake reloaded page</p>', 'lxml')
 
@@ -579,8 +836,8 @@ def test_refresh_follow_link():
 
     browser.refresh()
 
-    assert browser.get_url() == follow_url
-    assert browser.get_current_page() == reload_page
+    assert browser.url == follow_url
+    assert browser.page == reload_page
 
 
 def test_refresh_form_not_retained():
@@ -598,9 +855,10 @@ def test_refresh_form_not_retained():
 
     browser.refresh()
 
-    assert browser.get_url() == url
-    assert browser.get_current_page() == reload_page
-    assert browser.get_current_form() is None
+    assert browser.url == url
+    assert browser.page == reload_page
+    with pytest.raises(AttributeError, match="No form has been selected yet."):
+        browser.form
 
 
 def test_refresh_error():
@@ -616,5 +874,15 @@ def test_refresh_error():
         browser.refresh()
 
 
+def test_requests_session_and_cookies(httpbin):
+    """Check that the session object passed to the constructor of
+    StatefulBrowser is actually taken into account."""
+    s = requests.Session()
+    requests.utils.add_dict_to_cookiejar(s.cookies, {'key1': 'val1'})
+    browser = mechanicalsoup.StatefulBrowser(session=s)
+    resp = browser.get(httpbin + "/cookies")
+    assert resp.json() == {'cookies': {'key1': 'val1'}}
+
+
 if __name__ == '__main__':
     pytest.main(sys.argv)
